@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2013  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -29,12 +29,6 @@
 
 #include "connman.h"
 
-enum time_updates {
-	TIME_UPDATES_UNKNOWN = 0,
-	TIME_UPDATES_MANUAL  = 1,
-	TIME_UPDATES_AUTO    = 2,
-};
-
 enum timezone_updates {
 	TIMEZONE_UPDATES_UNKNOWN = 0,
 	TIMEZONE_UPDATES_MANUAL  = 1,
@@ -94,15 +88,79 @@ static enum timezone_updates string2timezone_updates(const char *value)
         return TIMEZONE_UPDATES_UNKNOWN;
 }
 
+static void clock_properties_load(void)
+{
+	GKeyFile *keyfile;
+	char *str;
+	enum time_updates time_value;
+	enum timezone_updates timezone_value;
+
+	keyfile = __connman_storage_load_global();
+	if (!keyfile)
+		return;
+
+	str = g_key_file_get_string(keyfile, "global", "TimeUpdates", NULL);
+
+	time_value = string2time_updates(str);
+	if (time_value != TIME_UPDATES_UNKNOWN)
+		time_updates_config = time_value;
+
+	g_free(str);
+
+	str = g_key_file_get_string(keyfile, "global", "TimezoneUpdates",
+			NULL);
+
+	timezone_value = string2timezone_updates(str);
+	if (timezone_value != TIMEZONE_UPDATES_UNKNOWN)
+		timezone_updates_config = timezone_value;
+
+	g_free(str);
+
+	g_key_file_free(keyfile);
+}
+
+static void clock_properties_save(void)
+{
+	GKeyFile *keyfile;
+	const char *str;
+
+	keyfile = __connman_storage_load_global();
+	if (!keyfile)
+		keyfile = g_key_file_new();
+
+	str = time_updates2string(time_updates_config);
+	if (str)
+		g_key_file_set_string(keyfile, "global", "TimeUpdates", str);
+	else
+		g_key_file_remove_key(keyfile, "global", "TimeUpdates", NULL);
+
+	str = timezone_updates2string(timezone_updates_config);
+	if (str)
+		g_key_file_set_string(keyfile, "global", "TimezoneUpdates",
+				str);
+	else
+		g_key_file_remove_key(keyfile, "global", "TimezoneUpdates",
+				NULL);
+
+	__connman_storage_save_global(keyfile);
+
+	g_key_file_free(keyfile);
+}
+
+enum time_updates __connman_clock_timeupdates(void)
+{
+	return time_updates_config;
+}
+
 static void append_timeservers(DBusMessageIter *iter, void *user_data)
 {
 	int i;
 	char **timeservers = __connman_timeserver_system_get();
 
-	if (timeservers == NULL)
+	if (!timeservers)
 		return;
 
-	for (i = 0; timeservers[i] != NULL; i++) {
+	for (i = 0; timeservers[i]; i++) {
 		dbus_message_iter_append_basic(iter,
 				DBUS_TYPE_STRING, &timeservers[i]);
 	}
@@ -121,7 +179,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBG("conn %p", conn);
 
 	reply = dbus_message_new_method_return(msg);
-	if (reply == NULL)
+	if (!reply)
 		return NULL;
 
 	dbus_message_iter_init_append(reply, &array);
@@ -136,16 +194,16 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	}
 
 	str = time_updates2string(time_updates_config);
-	if (str != NULL)
+	if (str)
 		connman_dbus_dict_append_basic(&dict, "TimeUpdates",
 						DBUS_TYPE_STRING, &str);
 
-	if (timezone_config != NULL)
+	if (timezone_config)
 		connman_dbus_dict_append_basic(&dict, "Timezone",
 					DBUS_TYPE_STRING, &timezone_config);
 
 	str = timezone_updates2string(timezone_updates_config);
-	if (str != NULL)
+	if (str)
 		connman_dbus_dict_append_basic(&dict, "TimezoneUpdates",
 						DBUS_TYPE_STRING, &str);
 
@@ -166,7 +224,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	DBG("conn %p", conn);
 
-	if (dbus_message_iter_init(msg, &iter) == FALSE)
+	if (!dbus_message_iter_init(msg, &iter))
 		return __connman_error_invalid_arguments(msg);
 
 	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
@@ -182,7 +240,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	type = dbus_message_iter_get_arg_type(&value);
 
-	if (g_str_equal(name, "Time") == TRUE) {
+	if (g_str_equal(name, "Time")) {
 		struct timeval tv;
 		dbus_uint64_t newval;
 
@@ -203,7 +261,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 		connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
 				CONNMAN_CLOCK_INTERFACE, "Time",
 				DBUS_TYPE_UINT64, &newval);
-	} else if (g_str_equal(name, "TimeUpdates") == TRUE) {
+	} else if (g_str_equal(name, "TimeUpdates")) {
 		const char *strval;
 		enum time_updates newval;
 
@@ -221,10 +279,11 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		time_updates_config = newval;
 
+		clock_properties_save();
 		connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
 				CONNMAN_CLOCK_INTERFACE, "TimeUpdates",
 				DBUS_TYPE_STRING, &strval);
-	} else if (g_str_equal(name, "Timezone") == TRUE) {
+	} else if (g_str_equal(name, "Timezone")) {
 		const char *strval;
 
 		if (type != DBUS_TYPE_STRING)
@@ -236,8 +295,8 @@ static DBusMessage *set_property(DBusConnection *conn,
 		dbus_message_iter_get_basic(&value, &strval);
 
 		if (__connman_timezone_change(strval) < 0)
-                        return __connman_error_invalid_arguments(msg);
-	} else if (g_str_equal(name, "TimezoneUpdates") == TRUE) {
+			return __connman_error_invalid_arguments(msg);
+	} else if (g_str_equal(name, "TimezoneUpdates")) {
 		const char *strval;
 		enum timezone_updates newval;
 
@@ -255,10 +314,11 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		timezone_updates_config = newval;
 
+		clock_properties_save();
 		connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
 				CONNMAN_CLOCK_INTERFACE, "TimezoneUpdates",
 				DBUS_TYPE_STRING, &strval);
-	} else if (g_str_equal(name, "Timeservers") == TRUE) {
+	} else if (g_str_equal(name, "Timeservers")) {
 		DBusMessageIter entry;
 		char **str = NULL;
 		GSList *list = NULL;
@@ -284,10 +344,10 @@ static DBusMessage *set_property(DBusConnection *conn,
 			dbus_message_iter_next(&entry);
 		}
 
-		if (list != NULL) {
+		if (list) {
 			str = g_new0(char *, count+1);
 
-			while (list != NULL) {
+			while (list) {
 				count--;
 				str[count] = list->data;
 				list = g_slist_delete_link(list, list);
@@ -296,7 +356,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		__connman_timeserver_system_set(str);
 
-		if (str != NULL)
+		if (str)
 			g_strfreev(str);
 
 		connman_dbus_property_changed_array(CONNMAN_MANAGER_PATH,
@@ -333,7 +393,7 @@ void __connman_clock_update_timezone(void)
 	g_free(timezone_config);
 	timezone_config = __connman_timezone_lookup();
 
-	if (timezone_config == NULL)
+	if (!timezone_config)
 		return;
 
 	connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
@@ -346,7 +406,7 @@ int __connman_clock_init(void)
 	DBG("");
 
 	connection = connman_dbus_get_connection();
-	if (connection == NULL)
+	if (!connection)
 		return -1;
 
 	__connman_timezone_init();
@@ -358,6 +418,7 @@ int __connman_clock_init(void)
 						clock_methods, clock_signals,
 						NULL, NULL, NULL);
 
+	clock_properties_load();
 	return 0;
 }
 
@@ -365,7 +426,7 @@ void __connman_clock_cleanup(void)
 {
 	DBG("");
 
-	if (connection == NULL)
+	if (!connection)
 		return;
 
 	g_dbus_unregister_interface(connection, CONNMAN_MANAGER_PATH,
